@@ -1,30 +1,37 @@
-from collections import defaultdict
 from contextlib import contextmanager
 from enum import Enum
+from functools import wraps
 from itertools import chain
 from typing import (  # type: ignore
     AbstractSet,
     Any,
+    Callable,
     Collection,
-    Dict,
     Iterable,
     List,
     Mapping,
     Optional,
-    Pattern,
     Sequence,
-    Set,
     Tuple,
     Type,
+    TypeVar,
     Union,
     cast,
 )
 
+from dataclasses import Field
+
 from apischema import settings
 from apischema.conversions.utils import Conversions
 from apischema.conversions.visitor import Conv
-from apischema.dataclasses import is_dataclass, replace
-from apischema.dataclasses.cache import Field, FieldKind
+from apischema.dataclass_utils import (
+    get_alias,
+    get_default,
+    has_default,
+    is_dataclass,
+    is_required,
+)
+from apischema.dataclasses import replace
 from apischema.json_schema.constraints import (
     ArrayConstraints,
     Constraints,
@@ -40,18 +47,18 @@ from apischema.json_schema.generation.visitor import (
 )
 from apischema.json_schema.patterns import infer_pattern
 from apischema.json_schema.refs import get_ref, schema_ref
-from apischema.json_schema.schema import (
-    Schema,
-    get_schema,
-    merge_schema,
-    serialize_schema,
-)
+from apischema.json_schema.schema import Schema, get_schema, merge_schema
 from apischema.json_schema.types import JsonSchema, JsonType, json_schema
 from apischema.json_schema.versions import JsonSchemaVersion, RefFactory
-from apischema.metadata.keys import is_aggregate_field
+from apischema.metadata.keys import (
+    MERGED_METADATA,
+    PROPERTIES_METADATA,
+    SCHEMA_METADATA,
+    check_metadata,
+)
 from apischema.serialization import serialize
 from apischema.types import AnyType
-from apischema.utils import get_default, is_hashable
+from apischema.utils import is_hashable
 
 constraint_by_type = {
     int: NumberConstraints,
@@ -60,30 +67,23 @@ constraint_by_type = {
 }
 
 
-def check_constraints(schema: Optional[Schema], expected: Type[Constraints]):
-    if schema is not None and schema.constraints is not None:
-        if not isinstance(schema.constraints, expected):
-            raise TypeError(
-                f"Bad constraints: expected {expected.__name__}"
-                f" found {type(schema.constraints).__name__}"
-            )
+Method = TypeVar("Method", bound=Callable[..., JsonSchema])
 
 
-def check_no_constraints(schema: Optional[Schema], type_):
-    if schema is not None and schema.constraints is not None:
-        raise TypeError(f"{type} cannot have constraints")
+def with_schema(method: Method) -> Method:
+    @wraps(method)
+    def wrapper(self: "SchemaBuilder", *args, **kwargs):
+        if self.schema is None:
+            return method(self, *args, **kwargs)
+        elif self.schema.override:
+            return JsonSchema(**self.schema.as_dict())
+        else:
+            return JsonSchema(method(self, *args, **kwargs), **self.schema.as_dict())
+
+    return cast(Method, wrapper)
 
 
-def _with_schema(schema: Optional[Schema], json_schema: JsonSchema) -> JsonSchema:
-    if schema is None:
-        return json_schema
-    elif schema.override:
-        return JsonSchema(**serialize_schema(schema))
-    else:
-        return JsonSchema(json_schema, **serialize_schema(schema))
-
-
-class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
+class SchemaBuilder(SchemaVisitor[Conv, JsonSchema]):
     def __init__(
         self,
         conversions: Optional[Conversions],
@@ -95,13 +95,24 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
         self.ref_factory = ref_factory
         self.refs = refs
         self.ignore_first_ref = ignore_first_ref
+        self.schema: Optional[Schema] = None
 
-    def _ref_schema(self, ref: str, schema: Optional[Schema]) -> JsonSchema:
-        return _with_schema(schema, JsonSchema({"$ref": self.ref_factory(ref)}))
+    def _check_constraints(self, expected: Type[Constraints]):
+        if self.schema is not None and self.schema.constraints is not None:
+            if not isinstance(self.schema.constraints, expected):
+                raise TypeError(
+                    f"Bad constraints: expected {expected.__name__}"
+                    f" found {type(self.schema.constraints).__name__}"
+                )
 
-    def _annotated(
-        self, cls: AnyType, annotations: Sequence[Any], schema: Optional[Schema]
-    ) -> JsonSchema:
+    def _merge_schema(self, schema: Optional[Schema]):
+        self.schema = merge_schema(schema, self.schema)
+
+    @with_schema
+    def _ref_schema(self, ref: str) -> JsonSchema:
+        return JsonSchema({"$ref": self.ref_factory(ref)})
+
+    def annotated(self, cls: AnyType, annotations: Sequence[Any]) -> JsonSchema:
         for annotation in reversed(annotations):
             if isinstance(annotation, schema_ref):
                 annotation.check_type(cls)
@@ -110,62 +121,47 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
                         self.ignore_first_ref = False
                     else:
                         assert isinstance(annotation.ref, str)
-                        return self._ref_schema(annotation.ref, schema)
+                        return self._ref_schema(annotation.ref)
                 ref = annotation.ref
                 if not isinstance(ref, str):
                     raise ValueError("Annotated schema_ref can only be str")
             if isinstance(annotation, Schema):
-                schema = merge_schema(annotation, schema)
-        return self.visit(cls, schema)
+                self._merge_schema(annotation)
+        return self.visit_with_schema(cls, self.schema)
 
-    def any(self, schema: Optional[Schema]) -> JsonSchema:
-        check_no_constraints(schema, "Any")
-        return _with_schema(schema, JsonSchema())
+    @with_schema
+    def any(self) -> JsonSchema:
+        return JsonSchema()
 
-    def collection(
-        self, cls: Type[Iterable], value_type: AnyType, schema: Optional[Schema]
-    ) -> JsonSchema:
-        check_constraints(schema, ArrayConstraints)
-        return _with_schema(
-            schema,
-            json_schema(
-                type=JsonType.ARRAY,
-                items=self.visit(value_type),
-                uniqueItems=issubclass(cls, AbstractSet),
-            ),
+    @with_schema
+    def collection(self, cls: Type[Iterable], value_type: AnyType) -> JsonSchema:
+        self._check_constraints(ArrayConstraints)
+        return json_schema(
+            type=JsonType.ARRAY,
+            items=self.visit(value_type),
+            uniqueItems=issubclass(cls, AbstractSet),
         )
 
-    @staticmethod
-    def _override_arg(cls: AnyType, schema: Optional[Schema]) -> Optional[Schema]:
-        return merge_schema(get_schema(cls), schema)
+    def _visit_field_(self, field: Field, field_type: AnyType):
+        if SCHEMA_METADATA in field.metadata:
+            schema: Schema = field.metadata[SCHEMA_METADATA]
+            if schema.annotations is not None and schema.annotations is ...:
+                if not has_default(field):
+                    raise TypeError("Invalid ... without field default")
+                try:
+                    default = serialize(get_default(field))
+                except Exception:
+                    pass
+                else:
+                    annotations = replace(schema.annotations, default=default)
+                    schema = replace(schema, annotations=annotations)
+            return self.visit_with_schema(field_type, schema)
+        else:
+            return self.visit(field_type)
 
-    def _field_schema(self, field: Field) -> Schema:
-        annotations = field.annotations
-        if annotations is not None and annotations.default is ...:
-            if not field.default:
-                raise TypeError("Invalid ... without field default")
-            try:
-                default = serialize(get_default(field.base_field))
-            except Exception:
-                pass
-            else:
-                annotations = replace(annotations, default=default)
-        return Schema(annotations, field.constraints)
-
-    def _visit_field(self, field: Field) -> JsonSchema:
-        result = self._field_visit(field, self._field_schema(field))
-        if not is_aggregate_field(field.base_field):
-            result = json_schema(
-                readOnly=field.kind == FieldKind.NO_INIT,
-                writeOnly=field.kind == FieldKind.INIT,
-                **result,
-            )
-        return result
-
-    def _properties_schema(self, field: Field) -> JsonSchema:
-        props_schema = JsonSchema()
+    def _properties_schema(self, field: Field, field_type: AnyType) -> JsonSchema:
         with self._without_ref():
-            props_schema = self._visit_field(field)
+            props_schema = self._visit_field(field, field_type)
         if not props_schema.get("type") == JsonType.OBJECT:
             raise TypeError("properties field must have an 'object' type")
         if "patternProperties" in props_schema:
@@ -183,49 +179,51 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
                 pass
         return JsonSchema()
 
-    _required: bool
-
-    def dataclass(self, cls: Type, schema: Optional[Schema]) -> JsonSchema:
+    @with_schema
+    def dataclass(
+        self,
+        cls: Type,
+        types: Mapping[str, AnyType],
+        fields: Sequence[Field],
+        init_vars: Sequence[Field],
+    ) -> JsonSchema:
         assert is_dataclass(cls)
-        check_constraints(schema, ObjectConstraints)
+        self._check_constraints(ObjectConstraints)
         properties = {}
         required: List[str] = []
-        (
-            fields,
-            merged_fields,
-            pattern_fields,
-            additional_field,
-        ) = self._dataclass_fields(cls)
-        dependent: Dict[str, Set[str]] = defaultdict(set)
-        for field in fields:
-            properties[field.alias] = self._visit_field(field)
-            if not field.default:
-                required.append(field.alias)
-            for req in self._required_by(field):
-                dependent[req].add(field.alias)
-        dependent_required = {
-            field: sorted(dependent[field]) for field in sorted(dependent)
-        }
-
-        merged_schemas = [self._visit_field(field) for _, field in merged_fields]
-        pattern_properties = {
-            cast(Pattern, pattern)
-            if pattern is not ...
-            else infer_pattern(field): self._properties_schema(field)
-            for pattern, field in pattern_fields
-        }
-        additional_properties: Union[bool, JsonSchema]
-        if additional_field:
-            additional_properties = self._properties_schema(additional_field)
-        else:
-            additional_properties = settings.additional_properties
+        merged_schemas = []
+        pattern_properties = {}
+        additional_properties: Union[bool, JsonSchema] = settings.additional_properties
+        for field in self._dataclass_fields(cls, fields, init_vars):
+            metadata = check_metadata(field)
+            field_type = types[field.name]
+            if MERGED_METADATA in metadata:
+                merged_schemas.append(self._visit_field(field, field_type))
+            elif PROPERTIES_METADATA in metadata:
+                pattern = metadata[PROPERTIES_METADATA]
+                properties_schema = self._properties_schema(field, field_type)
+                if pattern is None:
+                    additional_properties = properties_schema
+                elif pattern is ...:
+                    pattern_properties[infer_pattern(field_type)] = properties_schema
+                else:
+                    pattern_properties[pattern] = properties_schema
+            else:
+                alias = get_alias(field)
+                properties[alias] = json_schema(
+                    readOnly=not field.init,
+                    writeOnly=field in init_vars,
+                    **self._visit_field(field, field_type),
+                )
+                if is_required(field):
+                    required.append(alias)
         result = json_schema(
             type=JsonType.OBJECT,
             properties=properties,
             required=required,
             additionalProperties=additional_properties,
             patternProperties=pattern_properties,
-            dependentRequired=dependent_required,
+            dependentRequired=self._dependent_required(cls),
         )
         if merged_schemas:
             result = json_schema(
@@ -233,16 +231,15 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
                 allOf=[result, *merged_schemas],
                 unevaluatedProperties=False,
             )
-        return _with_schema(schema, result)
+        return result
 
-    def enum(self, cls: Type[Enum], schema: Optional[Schema]) -> JsonSchema:
-        check_no_constraints(schema, "Enum")
+    def enum(self, cls: Type[Enum]) -> JsonSchema:
         if len(cls) == 0:
             raise TypeError("Empty enum")
-        return self.literal(list(cls), schema)
+        return self.literal(list(cls))
 
-    def literal(self, values: Sequence[Any], schema: Optional[Schema]) -> JsonSchema:
-        check_no_constraints(schema, "Literal")
+    @with_schema
+    def literal(self, values: Sequence[Any]) -> JsonSchema:
         if not values:
             raise TypeError("Empty Literal")
         types = sorted(
@@ -253,116 +250,96 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
         )
         # Mypy issue
         type_: Any = types[0] if len(types) == 1 else types
-        return _with_schema(
-            schema,
-            json_schema(type_=type_, enum=values)
-            if len(values) != 1
-            else json_schema(type=type_, const=values[0]),
-        )
+        if len(values) == 1:
+            return json_schema(type=type_, const=values[0])
+        else:
+            return json_schema(type=type_, enum=values)
 
+    @with_schema
     def mapping(
         self,
         cls: Type[Mapping],
         key_type: AnyType,
         value_type: AnyType,
-        schema: Optional[Schema],
     ) -> JsonSchema:
-        check_constraints(schema, ObjectConstraints)
+        self._check_constraints(ObjectConstraints)
         with self._without_ref():
             key = self.visit(key_type)
         if key["type"] != JsonType.STRING:
             raise ValueError("Mapping types must string-convertible key")
         value = self.visit(value_type)
-        return _with_schema(
-            schema,
-            json_schema(type=JsonType.OBJECT, patternProperties={key["pattern"]: value})
-            if "pattern" in key
-            else json_schema(type=JsonType.OBJECT, additionalProperties=value),
-        )
+        if "pattern" in key:
+            return json_schema(
+                type=JsonType.OBJECT, patternProperties={key["pattern"]: value}
+            )
+        else:
+            return json_schema(type=JsonType.OBJECT, additionalProperties=value)
 
+    @with_schema
     def named_tuple(
         self,
         cls: Type[Tuple],
         types: Mapping[str, AnyType],
         defaults: Mapping[str, Any],
-        schema: Optional[Schema],
     ) -> JsonSchema:
-        check_constraints(schema, ObjectConstraints)
-        return _with_schema(
-            schema,
-            json_schema(
-                type=JsonType.OBJECT,
-                properties={key: self.visit(key) for key, cls in types.items()},
-                required=sorted(types.keys() - defaults.keys()),
-                additionalProperties=settings.additional_properties,
-            ),
+        self._check_constraints(ObjectConstraints)
+        return json_schema(
+            type=JsonType.OBJECT,
+            properties={key: self.visit(key) for key, cls in types.items()},
+            required=sorted(types.keys() - defaults.keys()),
+            additionalProperties=settings.additional_properties,
         )
 
-    def new_type(
-        self, cls: AnyType, super_type: AnyType, schema: Optional[Schema]
-    ) -> JsonSchema:
-        return self.visit(super_type, schema)
+    def new_type(self, cls: Type, super_type: AnyType) -> JsonSchema:
+        return self.visit_with_schema(super_type, self.schema)
 
-    def primitive(self, cls: Type, schema: Optional[Schema]) -> JsonSchema:
+    @with_schema
+    def primitive(self, cls: Type) -> JsonSchema:
         if cls in constraint_by_type:
-            check_constraints(schema, constraint_by_type[cls])
-        return _with_schema(schema, JsonSchema(type=JsonType.from_type(cls)))
+            self._check_constraints(constraint_by_type[cls])
+        return JsonSchema(type=JsonType.from_type(cls))
 
-    def subprimitive(
-        self, cls: Type, superclass: Type, schema: Optional[Schema]
-    ) -> JsonSchema:
-        return self.primitive(superclass, schema)
+    def subprimitive(self, cls: Type, superclass: Type) -> JsonSchema:
+        return self.primitive(superclass)
 
-    def tuple(self, types: Sequence[AnyType], schema: Optional[Schema]) -> JsonSchema:
-        check_constraints(schema, ArrayConstraints)
-        if schema is not None and schema.constraints is not None:
-            assert isinstance(schema.constraints, ArrayConstraints)
+    @with_schema
+    def tuple(self, types: Sequence[AnyType]) -> JsonSchema:
+        self._check_constraints(ArrayConstraints)
+        if self.schema is not None and self.schema.constraints is not None:
+            assert isinstance(self.schema.constraints, ArrayConstraints)
             if (
-                schema.constraints.max_items is not None
-                or schema.constraints.min_items is not None
+                self.schema.constraints.max_items is not None
+                or self.schema.constraints.min_items is not None
             ):
                 raise TypeError("Tuple cannot have min_items/max_items constraints")
-        return _with_schema(
-            schema,
-            json_schema(
-                type=JsonType.ARRAY,
-                items=[self.visit(cls) for cls in types],
-                minItems=len(types),
-                maxItems=len(types),
-            ),
+        return json_schema(
+            type=JsonType.ARRAY,
+            items=[self.visit(cls) for cls in types],
+            minItems=len(types),
+            maxItems=len(types),
         )
 
+    @with_schema
     def typed_dict(
-        self,
-        cls: Type,
-        keys: Mapping[str, AnyType],
-        total: bool,
-        schema: Optional[Schema],
+        self, cls: Type, keys: Mapping[str, AnyType], total: bool
     ) -> JsonSchema:
-        check_constraints(schema, ObjectConstraints)
-        return _with_schema(
-            schema,
-            json_schema(
-                type=JsonType.OBJECT,
-                properties={key: self.visit(cls) for key, cls in keys.items()},
-                required=list(keys) if total else [],
-            ),
+        self._check_constraints(ObjectConstraints)
+        return json_schema(
+            type=JsonType.OBJECT,
+            properties={key: self.visit(cls) for key, cls in keys.items()},
+            required=list(keys) if total else [],
         )
 
-    def _union_arg(self, cls: AnyType, arg: Optional[Schema]) -> Optional[Schema]:
-        return None
-
-    def _union_result(
-        self, results: Sequence[JsonSchema], schema: Optional[Schema]
-    ) -> JsonSchema:
+    @with_schema
+    def _union_result(self, results: Sequence[JsonSchema]) -> JsonSchema:
         if len(results) == 1:
-            result = results[0]
+            return results[0]
         elif all(alt.keys() == {"type"} for alt in results):
             types = chain.from_iterable(
                 [res["type"]] if isinstance(res["type"], JsonType) else res["type"]
                 for res in results
             )
-            result = json_schema(type=list(types))
+            return json_schema(type=list(types))
         elif (
             len(results) == 2
             and all("type" in res for res in results)
@@ -375,26 +352,34 @@ class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
                         types = [types]  # type: ignore
                     if "null" not in types:
                         result = JsonSchema({**result, "type": [*types, "null"]})
-                    break
+                    return result
             else:
                 raise NotImplementedError()
         else:
-            result = json_schema(anyOf=results)
-        return _with_schema(schema, result)
+            return json_schema(anyOf=results)
 
-    def visit(self, cls: AnyType, schema: Optional[Schema] = None) -> JsonSchema:
-        if self._is_conversion(cls):
-            return self.visit_not_builtin(cls, merge_schema(get_schema(cls), schema))
-        if is_hashable(cls):
-            ref = get_ref(cls)
-            if ref in self.refs:
-                if self.ignore_first_ref:
-                    self.ignore_first_ref = False
-                else:
-                    assert isinstance(ref, str)
-                    return self._ref_schema(ref, schema)
-            schema = merge_schema(get_schema(cls), schema)
-        return super().visit(cls, schema)
+    def visit_with_schema(self, cls: AnyType, schema: Optional[Schema]) -> JsonSchema:
+        schema_save = self.schema
+        self.schema = schema
+        try:
+            if self._is_conversion(cls):
+                self._merge_schema(get_schema(cls))
+                return self.visit_not_builtin(cls)
+            if is_hashable(cls):
+                ref = get_ref(cls)
+                if ref in self.refs:
+                    if self.ignore_first_ref:
+                        self.ignore_first_ref = False
+                    else:
+                        assert isinstance(ref, str)
+                        return self._ref_schema(ref)
+                self._merge_schema(get_schema(cls))
+            return super().visit(cls)
+        finally:
+            self.schema = schema_save
+
+    def visit(self, cls: AnyType):
+        return self.visit_with_schema(cls, None)
 
     @contextmanager
     def _without_ref(self):
@@ -472,7 +457,8 @@ def _schema(
         all_refs = True
     version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
     refs = _export_refs([(cls, conversions)], builder, all_refs)
-    json_schema = builder(conversions, ref_factory, refs, False).visit(cls, schema)
+    visitor = builder(conversions, ref_factory, refs, False)
+    json_schema = visitor.visit_with_schema(cls, schema)
     if add_defs:
         defs = _refs_schema(builder, refs, ref_factory)
         if defs:
